/** \file DisplayOpenGL.cpp
	\brief Display implementation based on OpenGL renderer
*/

/* Copyright (C) 2008-2009 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <vcl.h>

#pragma hdrstop

#include "DisplayOpenGL.h"
#include "ValueStringFormatter.h"
#include "TimeCounter.h"
#include <stdio.h>
#include <math.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <algorithm>
#include <assert.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)


void DisplayOpenGL::SetVsync(bool enabled)
{
    typedef BOOL (WINAPI * PFNWGLSWAPINTERVALEXTPROC) (int interval);
    PFNWGLSWAPINTERVALEXTPROC wglSwapInterval;
    wglSwapInterval = (PFNWGLSWAPINTERVALEXTPROC) wglGetProcAddress("wglSwapIntervalEXT");
    //sets the number of vert retraces before redrawing
    if ( wglSwapInterval )
        wglSwapInterval(enabled ? 1 : 0);
}

void DisplayOpenGL::CreateBmpFont(void)					 
{
	HFONT	font;						// Windows Font ID
	HFONT	oldfont;
	base = glGenLists(96);				// Storage For 96 Characters

	font = CreateFont(	-12,			// Height Of Font
		0,								// Width Of Font
		0,								// Angle Of Escapement
		0,								// Orientation Angle
		FW_BOLD,							// Font Weight
		FALSE,							// Italic
		FALSE,							// Underline
		FALSE,							// Strikeout
		ANSI_CHARSET,					// Character Set Identifier
		OUT_TT_PRECIS,					// Output Precision
		CLIP_DEFAULT_PRECIS,			// Clipping Precision
		ANTIALIASED_QUALITY,			// Output Quality
		FF_DONTCARE|DEFAULT_PITCH,		// Family And Pitch
		"Courier New");					// Font Name
	oldfont = (HFONT)SelectObject(this->hdc, font);
	wglUseFontBitmaps(this->hdc, 32, 96, base);
	SelectObject(this->hdc, oldfont);
	DeleteObject(font);
}

void DisplayOpenGL::DestroyFont(void)
{
	glDeleteLists(base, 96);	// delete char list
}

GLvoid DisplayOpenGL::Print(const char *fmt, ...)
{
	char buf[128];
	va_list	ap;
	if (fmt == NULL)
		return;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf)-1, fmt, ap);
	va_end(ap);
	buf[sizeof(buf)-1] = NULL;

	glPushAttrib(GL_LIST_BIT);
	glListBase(base - 32);
	glCallLists(strlen(buf), GL_UNSIGNED_BYTE, buf);
    glPopAttrib();										
}


//---------------------------------------------------------------------------
DisplayOpenGL::DisplayOpenGL(void* hParent):
	iWidth(100), iHeight(100)
{
	//_control87(MCW_EM, MCW_EM);
    SetVsync(true);
    hdc = GetDC(hParent);
    SetPixelFormatDescriptor();
    hrc = wglCreateContext(hdc);
    wglMakeCurrent(hdc, hrc);
	SetupRC();
	CreateBmpFont();

	data = NULL;
}
//---------------------------------------------------------------------------

DisplayOpenGL::~DisplayOpenGL()
{
    //kolejnosc?
    ReleaseDC(0, hdc);
    wglMakeCurrent(hdc, NULL);
    wglDeleteContext(hrc);
}


void DisplayOpenGL::ChangeSize(int iWidth, int iHeight)
{
	fWidth = (float)iWidth / State().fZoomX;
	fHeight = (float)fRange / State().fZoomY;

	glViewport(0, 0, iWidth, iHeight);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
	glOrtho(State().fOffsetX,          	// left edge
		  State().fOffsetX+fWidth,     	// right edge
		  -fHeight+State().fOffsetY,   	// bottom edge
		  fHeight+State().fOffsetY,    	// top edge
		  0.0, 1.0);             		// near and far clipping, planes
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glDisable(GL_DEPTH_TEST);			// not using depth buffer
	// Displacement trick for exact pixelization
	// glTranslatef(0.375, 0.375, 0);

	this->iWidth = iWidth;
	this->iHeight = iHeight;
}

void DisplayOpenGL::SetMode(const enum E_DISPLAY_MODE m)
{
	mode = m;
	ChangeSize(iWidth, iHeight);
	Update(UPDATE_ALL);
}


#define MYGL_COLORS(color) GetRValue(color)/256.0, GetGValue(color)/256.0, GetBValue(color)/256.0
//---------------------------------------------------------------------------
void DisplayOpenGL::Update(enum E_UPDATE_LEVEL level)
{
	static TColor colorGrid = (TColor)0x303030;	// BGR
	static TColor colorLabel = (TColor)0x505050;
	static TColor colorCursor1 = (TColor)0x0000FF;
	static TColor colorCursor2 = (TColor)0xFF0000;

	glClear(GL_COLOR_BUFFER_BIT);

	glColor3f(MYGL_COLORS(colorGrid));
	glEnable(GL_LINE_STIPPLE);         /* dashed lines */
	glLineStipple(1, 0xF0F0);
	glBegin(GL_LINES);
	if (State().bGridY)
	{
		// horizontal grid
		for (int i=0; i<9; i++)
		{
			glVertex2f(State().fOffsetX, fHeight*(i-4)/5.0f + State().fOffsetY); // origin
			glVertex2f(fWidth+State().fOffsetX, fHeight*(i-4)/5.0f + State().fOffsetY); // end
		}
	}
	if (State().bGridX)
	{
		// vertical grid
		for (int i=1; i<=9; i++)
		{
			glVertex2f(fWidth*(i)/10.0f + State().fOffsetX, -fHeight + State().fOffsetY); // origin
			glVertex2f(fWidth*(i)/10.0f + State().fOffsetX, fHeight + State().fOffsetY); // end
		}
	}
	glEnd();
	glDisable(GL_LINE_STIPPLE);

	glColor3f(MYGL_COLORS(colorLabel));
	glColor3f(0.6f,0.6f,0.6f);
	std::string label;

	if (State().bGridY && State().bGridLabels)
	{
        // horizontal grid labels
		for (int i=-4; i<=4; i+=2)
		{
			glRasterPos2f(fWidth/40.0 + State().fOffsetX, fHeight*((float)i+0.07f)/5.0f + State().fOffsetY);
			float fVal = (fHeight*((float)i)/5.0f) + State().fOffsetY;
			label = ValueStringFormatter::VoltageToStr(fVal, 2);
			Print("%s", label.c_str());
		}
	}

	if (State().bGridX && State().bGridLabels)
	{
		// vertical grid labels
		for (int i=1; i<=9; i+=2)
		{
			glRasterPos2f(fWidth*((float)i+0.07f)/10.0f + State().fOffsetX, -fHeight*0.95f + State().fOffsetY);
			float fVal;
			if (mode == MODE_OSCILLOSCOPE || mode == MODE_RECORDER)
			{
				fVal = (fWidth*((float)i)/10.0f + State().fOffsetX) * fDivX;
				label = ValueStringFormatter::TimeToStr(fVal, 2);
			}
			else if (mode == MODE_SPECTRUM)
			{
				if (data && data->size())
				{
					fVal = (fWidth*((float)i)/10.0f + State().fOffsetX) / (fDivX*NearestPowerOfTwo(data->size())*2.0f);
					label = ValueStringFormatter::FreqToStr(fVal, 2);
				} else
				{
                	label = "";
				}
			}
			Print("%s", label.c_str());
		}
	}

	// plot
	if (data && data->size())
	{
		if (mode == MODE_OSCILLOSCOPE || mode == MODE_RECORDER)
		{
			glColor3f(0.9f,0.9f,0.9f);
#if 1
			glBegin(GL_LINE_STRIP);
			int i = std::max<float>(State().fOffsetX, 0.0f);
			float limit = std::min<float>(data->size(), fWidth + i);
			//TimeCounter tc("DisplayOpenGL update");
			for (float fHPos = i; fHPos<limit; fHPos+=1.0f)
			{
				glVertex2f(fHPos, (*data)[i++]);
			}
			glEnd();
#else
			// strange: this version runs slowe
            vertices.reserve(data->size() * 2);
			int i = std::max<float>(State().fOffsetX, 0.0f);
			float limit = std::min<float>(data->size(), fWidth + i);
			//TimeCounter tc("DisplayOpenGL update");
			int j=0;

			{
			//TimeCounter tc("Adding vertices");
			for (float fHPos = i; fHPos<limit; fHPos+=1.0f)
			{
				vertices[j++] = fHPos;
				vertices[j++] = (*data)[i++];
			}
			}

			glEnableClientState(GL_VERTEX_ARRAY);
			glVertexPointer(2, GL_FLOAT, 0, &vertices[0]);
			glDrawArrays(GL_LINE_STRIP, 0, j/2);
			glDisableClientState(GL_VERTEX_ARRAY);
#endif
		}
		else if (mode == MODE_SPECTRUM)
		{
			glColor3f(0.9f,0.9f,0.9f);
			int i = std::max<float>(State().fOffsetX, 0.0f);
			float limit = std::min<float>(data->size(), fWidth + i);
			glBegin(GL_LINES);
			for (float fHPos = i; fHPos<limit; fHPos+=1.0f)
			{
				glVertex2f(fHPos, 0.0f);
				glVertex2f(fHPos, (*data)[i++]);
			}
			glEnd();

			i = std::max<float>(State().fOffsetX, 0.0f);
			glBegin(GL_POINT);
			for (float fHPos = i; fHPos<limit; fHPos+=1.0f)
			{
				if ((*data)[i] == 0.0f)
				{
					glVertex2f(fHPos, 0.0f);
				}
				i++;
			}
			glEnd();
		}
	}

	// cursors
	glColor3f(MYGL_COLORS(colorCursor1));
	glBegin(GL_LINES);
	glVertex2f(State().fCursor1X-fWidth/(2*10*iWidth/iHeight), State().fCursor1Y);
	glVertex2f(State().fCursor1X+fWidth/(2*10*iWidth/iHeight), State().fCursor1Y);
	glVertex2f(State().fCursor1X, State().fCursor1Y-fHeight/(10));
	glVertex2f(State().fCursor1X, State().fCursor1Y+fHeight/(10));
	glEnd();

	glColor3f(MYGL_COLORS(colorCursor2));
	glBegin(GL_LINES);
	glVertex2f(State().fCursor2X-fWidth/(2*10*iWidth/iHeight), State().fCursor2Y);
	glVertex2f(State().fCursor2X+fWidth/(2*10*iWidth/iHeight), State().fCursor2Y);
	glVertex2f(State().fCursor2X, State().fCursor2Y-fHeight/10);
	glVertex2f(State().fCursor2X, State().fCursor2Y+fHeight/10);
	glEnd();	

	glFlush();
	SwapBuffers(hdc);
}
//---------------------------------------------------------------------------
void DisplayOpenGL::SetPixelFormatDescriptor()
{
	PIXELFORMATDESCRIPTOR pfd;
	ZeroMemory(&pfd, sizeof(pfd));					//ustalenie typu koloru, buforowania grafiki,
    pfd.nSize = sizeof(pfd);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 24;
    pfd.cDepthBits = 16;
    pfd.iLayerType = PFD_MAIN_PLANE;
    int PixelFormat = ChoosePixelFormat(hdc, &pfd);
    SetPixelFormat(hdc, PixelFormat, &pfd);
}
//---------------------------------------------------------------------------
void DisplayOpenGL::SetupRC(void)
{
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);	// background color
	glClear(GL_COLOR_BUFFER_BIT);
	glFlush();
}

float DisplayOpenGL::FPosToPhysicalX(float fPosX)
{
	float fVal;
	assert(mode == MODE_OSCILLOSCOPE || mode == MODE_SPECTRUM || mode == MODE_RECORDER);
	if (mode == MODE_OSCILLOSCOPE || mode == MODE_RECORDER)
		fVal = fPosX * fDivX;
	else if (mode == MODE_SPECTRUM)
		fVal = fPosX / (fDivX*NearestPowerOfTwo(data->size())*2.0f);	///< \todo depends on FFT samples, not data size
	return fVal;
}

float DisplayOpenGL::FPosToPhysicalY(float fPosY)
{
	return fPosY;
}

float DisplayOpenGL::IPosToFloatX(int x)
{
	float val;
	val = State().fOffsetX + ((float)(x+1)/float(iWidth)*fWidth);
	return val;
}

float DisplayOpenGL::IPosToFloatY(int y)
{
	float val;
	val = State().fOffsetY +
		((float)(iHeight-y-1)/float(iHeight)*fHeight*2) - fHeight;
	return val;
}


void DisplayOpenGL::SetCursor1(int x, int y)
{
	State().fCursor1X = IPosToFloatX(x);
	State().fCursor1Y = IPosToFloatY(y);
}

void DisplayOpenGL::SetCursor2(int x, int y)
{
	State().fCursor2X = IPosToFloatX(x);
	State().fCursor2Y = IPosToFloatY(y);
}

void DisplayOpenGL::PosToPhysical(int x, int y, float &fValX, float &fValY)
{
	fValX = FPosToPhysicalX(IPosToFloatX(x));
	fValY = FPosToPhysicalY(IPosToFloatY(y));
}

int DisplayOpenGL::GetCursorPhysical(enum E_CURSOR cursor, float &fValX, float &fValY)
{
	if (cursor == CURSOR1)
	{
		fValX = FPosToPhysicalX(State().fCursor1X);
		fValY = FPosToPhysicalY(State().fCursor1Y);
	}
	else if (cursor == CURSOR2)
	{
		fValX = FPosToPhysicalX(State().fCursor2X);
		fValY = FPosToPhysicalY(State().fCursor2Y);
	}
	else
		return -1;
	return 0;
}

void DisplayOpenGL::SetZoom(int iCenterX, int iCenterY, float fZoomX, float fZoomY)
{
	this->State().fZoomX = fZoomX;
	this->State().fZoomY = fZoomY;

	if (iCenterX == OFFSET_CLEAR) {
		this->State().fOffsetX = 0;
	} else if (iCenterX == OFFSET_LEAVE) {
	} else {
		this->State().fOffsetX = IPosToFloatX(iCenterX);
		fWidth = (float)iWidth / State().fZoomX;
		this->State().fOffsetX -= fWidth/2;
	}

	if (iCenterY == OFFSET_CLEAR) {
		this->State().fOffsetY = 0;
	} else if (iCenterY == OFFSET_LEAVE) {
	} else {
		this->State().fOffsetY = IPosToFloatY(iCenterY);
	}

	ChangeSize(this->iWidth, this->iHeight);
}

void DisplayOpenGL::SetOffsetY(float fVal)
{
	this->State().fOffsetY = fVal;
	ChangeSize(this->iWidth, this->iHeight);
}


void DisplayOpenGL::Move(int iStartX, int iStartY, int iStopX, int iStopY)
{
	float deltaX = IPosToFloatX(iStartX) - IPosToFloatX(iStopX);
	float deltaY = IPosToFloatY(iStartY) - IPosToFloatY(iStopY);
	this->State().fOffsetX += deltaX;
	this->State().fOffsetY += deltaY;
	ChangeSize(this->iWidth, this->iHeight);
}

void DisplayOpenGL::SetRange(float range)
{
	Display::SetRange(range);
	//spectrum.fOffsetY = 0.8 * range;
    ChangeSize(iWidth, iHeight);
}

std::auto_ptr<Graphics::TBitmap> DisplayOpenGL::CreateBitmap(void)
{
	std::auto_ptr<Graphics::TBitmap> bitmap(new Graphics::TBitmap());
	bitmap->PixelFormat = pf24bit;
	bitmap->Width = iWidth;
	bitmap->Height = iHeight;

	int ww = ((bitmap->Width * 24 + 31) & ~31) >> 3;
	unsigned char *pData = new unsigned char[ww * bitmap->Height];
	glReadPixels(0, 0, bitmap->Width, bitmap->Height, GL_RGB, GL_UNSIGNED_BYTE, pData);

	for(int y=0; y<bitmap->Height; y++)
	{
		Byte* ptr = (Byte*)bitmap->ScanLine[iHeight-y-1];
		for(int x=0; x<bitmap->Width; x++)
		{
			ptr[x*3] = pData[y*ww + x*3 + 2];
			ptr[x*3+1]= pData[y*ww + x*3 + 1];
			ptr[x*3+2]= pData[y*ww + x*3];
		}
	}
	delete[] pData;

	return bitmap;
}

